package org.colinvella.fancyfish.entity.fish;

import org.colinvella.fancyfish.FancyFishMod;
import org.colinvella.fancyfish.entity.ai.FishEntityAIFlopAround;
import org.colinvella.fancyfish.entity.ai.FishEntityAIMate;
import org.colinvella.fancyfish.entity.ai.FishEntityAISwimAround;
import org.colinvella.fancyfish.entity.ai.FishEntityAITempt;
import org.colinvella.fancyfish.item.CapturedFishItem;
import org.colinvella.fancyfish.item.FishScaleItem;
import org.colinvella.fancyfish.item.ModItemRegistry;
import org.colinvella.fancyfish.logger.ModLogger;

import net.minecraft.block.Block;
import net.minecraft.entity.SharedMonsterAttributes;
import net.minecraft.entity.SharedMonsterAttributes;
import net.minecraft.entity.passive.EntityAnimal;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Items;
import net.minecraft.init.Items;
import net.minecraft.init.Items;
import net.minecraft.item.Item;
import net.minecraft.item.Item;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.item.ItemStack;
import net.minecraft.util.BlockPos;
import net.minecraft.util.DamageSource;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.world.World;

public abstract class FishEntity extends EntityAnimal
{
	public enum Species
	{
		ClownFish,
		RegalTang,
		MoorishIdol,
		PufferFish,
		RoyalGramma,
		YellowTang,
		DamselFish,
		Blenny,
		MoonFish,
		Flounder,
		GoldFish,
		AngelFish,
		Anthias,
		ButterflyFish,
		CardinalFish,
		Guppy
	}
	
    public FishEntity(World worldIn, Species species, double swimmingSpeed, FishScaleItem[] droppableScales)
    {
    	super(worldIn);
    	
    	if (logger == null)
    		logger = FancyFishMod.getLogger();

    	this.species = species;
    	this.swimmingSpeed = swimmingSpeed;
    	this.droppableScales = droppableScales;
    	
        this.setSize(0.25f, 0.25f);
        
        this.tasks.addTask(0, new FishEntityAIFlopAround(this));
        this.tasks.addTask(1, new FishEntityAIMate(this));
        this.tasks.addTask(2, new FishEntityAITempt(this, ModItemRegistry.FishFood, false));
        this.tasks.addTask(3, new FishEntityAISwimAround(this));
        
        logger.debug("Fish entity " + this.species + " spawned");
    }

	public Species getSpecies()
    {
    	return this.species;
    }
	
	public double getSwimmingSpeed()
	{
		return swimmingSpeed;
	}
	
	public abstract CapturedFishItem getCapturedFishItem();
 
    @Override
    public boolean isBreedingItem(ItemStack itemStack)
    {
    	if (itemStack == null)
    		return false;
    	return itemStack.getItem() == ModItemRegistry.FishFood;
    }
    
    /**
     * returns if this entity triggers Block.onEntityWalking on the blocks they walk on. used for spiders and wolves to
     * prevent them from trampling crops
     */
    @Override
    protected boolean canTriggerWalking()
    {
        return false;
    }

    @Override
    public boolean canBreatheUnderwater()
    {
    	return true;
    }
    
    @Override
    protected boolean canDespawn()
    {
    	// want fish to stay in world like other passive animals
    	return false;
    }
    
    @Override
    public boolean isPushedByWater()
    {
    	return false;
    }
    
    @Override
    public void onEntityUpdate()
    {
		//logger.debug(this.posX + ", " + this.posY + ", " + this.posZ);

    	// handle drowning when outside water
        int air = this.getAir();
        super.onEntityUpdate();

        if (this.isEntityAlive() && !this.isInWater())
        {
            --air;
            this.setAir(air);

            if (this.getAir() == -20)
            {
                this.setAir(0);
                this.attackEntityFrom(DamageSource.drown, 10.0F);
            }
        }
        else
        {
            this.setAir(300);
        }
    }
    
    @Override public void onDeath(DamageSource cause)
    {
    	logger.debug("Fish entity died!");
    	super.onDeath(cause);
    };
    
    @Override
    public void onLivingUpdate()
    {
        super.onLivingUpdate();
             
        if (this.worldObj.isRemote)
        {
            // if dipping, release bubbles
    		if (Math.abs(motionX) < 0.01 && Math.abs(motionZ) < 0.01 && motionY < -0.01)
    		{
    			this.worldObj.spawnParticle(EnumParticleTypes.WATER_BUBBLE, posX, posY, posZ, 0, 1, 0);
    			
    			// play sound
    			//playSound("mob.fish.bubbles", getSoundVolume(), getSoundPitch());
    		}
        }
        else // !remote
        {
            if (this.inWater)
            {
            	// blend motion towards target motion for smooth turns
                this.motionX += (this.targetMotionX - this.motionX) * 0.1;
                this.motionY += (this.targetMotionY - this.motionY) * 0.1;
                this.motionZ += (this.targetMotionZ - this.motionZ) * 0.1;
            }
            else
            {
            	// blend motion towards target motion for smooth turns
                this.motionX += (this.targetMotionX - this.motionX) * 0.1;
                this.motionZ += (this.targetMotionZ - this.motionZ) * 0.1;                
                // simulates gravity (I assume!)
                this.motionY -= 0.08D;
                this.motionY *= 0.9800000190734863D;
            }
                        
            // make entity face direction of motion in XZ plane
            this.rotationYaw = -(float)Math.atan2(this.motionX, this.motionZ) * RAD_TO_DEG;
            
            // set fish pitch (angle up or down)
            double xzSpeed = Math.sqrt(this.motionX * this.motionX + this.motionZ * this.motionZ);
            this.rotationPitch = -(float)Math.asin(this.motionY / xzSpeed) * RAD_TO_DEG;;
        }      
    }
    
    @Override
    protected void applyEntityAttributes()
    {
        super.applyEntityAttributes();
        this.getEntityAttribute(SharedMonsterAttributes.maxHealth).setBaseValue(10.0D);
        this.getEntityAttribute(SharedMonsterAttributes.movementSpeed).setBaseValue(0.1);
    }
        
    /**
     * Returns the sound this mob makes while it's alive.
     */
    @Override
    protected String getLivingSound()
    {
        return FancyFishMod.MOD_ID + ":mob.fish.Living";
    }

    /**
     * Returns the sound this mob makes when it is hurt.
     */
    @Override
    protected String getHurtSound()
    {
        //return "mob.rabbit.hurt";
        return FancyFishMod.MOD_ID + ":mob.fish.Hurt";
    }

    /**
     * Returns the sound this mob makes on death.
     */
    @Override
    protected String getDeathSound()
    {
        //return "mob.cow.hurt";
        return FancyFishMod.MOD_ID + ":mob.fish.Death";
    }

    @Override
    protected void playStepSound(BlockPos blockPos, Block block)
    {
        this.playSound(FancyFishMod.MOD_ID + ":mob.fish.Living", getSoundVolume(), getSoundPitch());
    }

    /**
     * Returns the volume for the sounds this mob makes.
     */
    @Override
    protected float getSoundVolume()
    {
        return 0.4F;
    }

    // return a fish scale at random from the list of droppable scale colours
    @Override
    protected Item getDropItem()
    {
    	// default item in case no droppable scales defined
    	if (this.droppableScales == null)
    		return Items.leather;
    	
    	int choices = this.droppableScales.length;
    	
    	// default item in case no droppable scales array is empty
    	if (choices == 0)
    		return Items.leather;
    	
    	return this.droppableScales[this.rand.nextInt(choices)];
    }

    /**
     * Drop 0-2 items of this living's type
     * Rewritten so each item is chosen randomly per iteration to get mix of scale colours
     */
    @Override
    protected void dropFewItems(boolean recentlyHit, int lootEnchantmentLevel)
    {
        int itemCount = this.rand.nextInt(3) + this.rand.nextInt(1 + lootEnchantmentLevel);

        while (itemCount-- > 0)
        {
            this.dropItem(getDropItem(), 1);
        }
    }

    /**
     * Called when a player interacts with a mob. e.g. gets milk from a cow, gets into the saddle on a pig.
     */
    @Override
    public boolean interact(EntityPlayer player)
    {
        ItemStack itemstack = player.inventory.getCurrentItem();

        if (itemstack != null && itemstack.getItem() == Items.bucket && !player.capabilities.isCreativeMode)
        {
            if (itemstack.stackSize-- == 1)
            {
                player.inventory.setInventorySlotContents(player.inventory.currentItem, new ItemStack(Items.milk_bucket));
            }
            else if (!player.inventory.addItemStackToInventory(new ItemStack(Items.milk_bucket)))
            {
                player.dropPlayerItemWithRandomChoice(new ItemStack(Items.milk_bucket, 1, 0), false);
            }

            return true;
        }
        else
        {
            return super.interact(player);
        }
    }

    @Override
    public float getEyeHeight()
    {
        return this.height;
    }
    
    @Override
    public boolean getCanSpawnHere()
    {
    	return super.getCanSpawnHere();
    }
    
    @Override
    public void moveEntityWithHeading(float strafe, float forward)
    {
    	if (this.isInWater())
    	{
    		this.moveFlying(strafe, forward, 0.1f);
    		this.moveEntity(motionX, motionY, motionZ);
    		
    		// slow down
    		this.targetMotionX *= 0.99d;
    		this.targetMotionY *= 0.99d;
    		this.targetMotionZ *= 0.99d;
    	}
    	else
    	{
    		this.moveEntity(motionX, motionY, motionZ);
    	}
    }

    public void setTargetMotion(double motionX, double motionY, double motionZ)
    {
        this.targetMotionX = motionX;
        this.targetMotionY = motionY;
        this.targetMotionZ = motionZ;
    }
    
    public boolean hasTargetMotion()
    {
        return this.targetMotionX != 0.0 || this.targetMotionY != 0.0 || this.targetMotionZ != 0.0;
    }
        
    // species
    private Species species;
    
    // swimming speed
    private double swimmingSpeed;
    
    // colours of droppable scales when fish dies
    private FishScaleItem[] droppableScales;
    
    // target motion
    private double targetMotionX;
    private double targetMotionY;
    private double targetMotionZ;

	// protected logger reference
	protected static ModLogger logger;
		
	// radian to degree conversion factor
	private static final float RAD_TO_DEG = 180.0f / (float)Math.PI;
}